/* SameGame for PalmOS version 1.0
 *
 * PalmOS implementation of the popular game commonly known as SameGame
 *
 * written by Hein Venter
 * Feb 1998
 */

#pragma pack(2)
#include <Common.h>
#include <System/SysAll.h>
#include <SysUtils.h>
#include <TimeMgr.h>
#include <StringMgr.h>
#include <MemoryMgr.h>
#include <UI/UIAll.h>
#include "SameGame.h"

static int StartApplication(void);
static void EventLoop(void);
static void StopApplication(void);
static Boolean gameform(EventPtr event);
static Boolean infoform(EventPtr event);
static Boolean OpenDatabase(void);
static void Select(int index);
static void GetInfo(void);
static void Tumble(void);
static void GenerateArray(void);
static void DrawBitmap(int index);
static void DrawGameScreen(void);
static void UpdateScore(void);

#define SameGameAppID    'SmGm'
#define SameGameDBType   'Data'

FieldPtr	fieldptr_fform;
FieldPtr	fieldptr_fpoints;
FieldPtr	fieldptr_fscore;
FieldPtr	fieldptr_ft;
FieldPtr	fieldptr_f1;
FieldPtr	fieldptr_f2;
FieldPtr	fieldptr_f3;
FieldPtr	fieldptr_f4;
FieldPtr	fieldptr_f5;

DmOpenRef	SameGameDB;
char        SameGameDBName[] = "SameGameDB";
int         score=0, points=0, formv=0;
int			lastpoints=0;
int			topscore=0;
char        gamearray[400];
char		undoarray[320];
int		    selprev;

DWord  PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
  int error;

  if (cmd == sysAppLaunchCmdNormalLaunch)
  {
    error = StartApplication();	
    if (error) return error;
    EventLoop();	
    StopApplication ();	
  }
  return 0;
}

static int StartApplication(void)
{
  int error;

  error = OpenDatabase();
  if (error) return error;
  FrmGotoForm(formID_game);
}

static Boolean OpenDatabase(void)
{
  UInt          index = 0;
  VoidHand      RecHandle;
  Ptr           RecPointer;
  char			sc[5];

  SameGameDB = DmOpenDatabaseByTypeCreator(SameGameDBType, SameGameAppID, dmModeReadWrite);
  if (!SameGameDB) 
  {

	// If the database doesn't exist, create a new one

	if (DmCreateDatabase(0, SameGameDBName, SameGameAppID, SameGameDBType, false)) return 1;
    SameGameDB = DmOpenDatabaseByTypeCreator(SameGameDBType, SameGameAppID, dmModeReadWrite);
    RecHandle = DmNewRecord(SameGameDB, &index, 400);
    RecPointer = MemHandleLock(RecHandle);
    DmWrite(RecPointer, 0, &gamearray, 390);
    MemPtrUnlock(RecPointer); 
    DmReleaseRecord(SameGameDB, index, true);
    GenerateArray();
  } 
  else
  {

	// Get the game state from the database

	RecHandle = DmGetRecord(SameGameDB, 0);
	RecPointer = MemHandleLock(RecHandle);
	StrNCopy(gamearray, RecPointer, 320);
    StrNCopy(undoarray, RecPointer, 320);
	StrNCopy(gamearray+350, RecPointer+350, 5);
	StrNCopy(gamearray+360, RecPointer+360, 5);
	MemPtrUnlock(RecPointer); 
	DmReleaseRecord(SameGameDB, index, 0);
	score=StrAToI(gamearray+350);
	topscore=StrAToI(gamearray+360);
  }
  return 0;
}

static void EventLoop(void)
{
  short err;
  int formID;
  FormPtr form;
  EventType event;

  do
  {
	EvtGetEvent(&event, -1);
    if (SysHandleEvent(&event)) continue;
    if (MenuHandleEvent((void *)0, &event, &err)) continue;
	if (event.eType == frmLoadEvent)
    {
      formID = event.data.frmLoad.formID;
      form = FrmInitForm(formID);
      FrmSetActiveForm(form);
      switch (formID) 
      {
        case formID_game:
			FrmSetEventHandler(form, (FormEventHandlerPtr) gameform);
			break;
		case formID_info:
			FrmSetEventHandler(form, (FormEventHandlerPtr) infoform);
			break;
      }
    }
    FrmDispatchEvent(&event);
  } while(event.eType != appStopEvent);
}

static void StopApplication(void)
{
	VoidHand      RecHandle;
	Ptr           RecPointer;
	int				i;

	for (i=0; i<320; i++)
	{
		if (gamearray[i]==6) gamearray[i]=selprev;
	}

	// Save the state of the game (all contained in 'gamearray'
	// This includes the current game state, score, and top score

	RecHandle = DmGetRecord(SameGameDB, 0);
    RecPointer = MemHandleLock(RecHandle);
    DmWrite(RecPointer, 0, &gamearray, 390);
    MemPtrUnlock(RecPointer); 
    DmReleaseRecord(SameGameDB, 0, true);
    DmCloseDatabase(SameGameDB);
}

static Boolean gameform(EventPtr event)
{
  FormPtr		form;
  int			handled = 0;
  EventType		newEvent;
  VoidHand      RecHandle;
  Ptr           RecPointer;
  int			penx, peny;
  int			row, column, index;

  switch (event->eType) 
  {
    case frmOpenEvent:
      form = FrmGetActiveForm();
      fieldptr_fform = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_fform));
      fieldptr_fpoints = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_fpoints));
      fieldptr_fscore = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_fscore));
      FrmDrawForm(form);
      DrawGameScreen();
	  UpdateScore();
      handled = 1;
      break;
	  
    case menuEvent:
      switch (event->data.menu.itemID)
	  {
        case menuitemID_newgame:
			GenerateArray();
			DrawGameScreen();
			UpdateScore();
			break;
		case menuitemID_info:
			FrmGotoForm(formID_info);
			break;
		case menuitemID_undo:
			StrNCopy(gamearray, undoarray, 320);
			DrawGameScreen();
			score=score-lastpoints;
			UpdateScore();
			break;
		case menuitemID_replay:
			RecHandle = DmGetRecord(SameGameDB, 0);
			RecPointer = MemHandleLock(RecHandle);
			StrNCopy(gamearray, RecPointer, 320);
			StrNCopy(gamearray+350, RecPointer+350, 5);
			score=StrAToI(gamearray+350);
			MemPtrUnlock(RecPointer); 
			DmReleaseRecord(SameGameDB, 0, 0);
			DrawGameScreen();
			UpdateScore();
			break;
        case menuitemID_about:
			FrmAlert(alertID_about);
			break;
        case menuitemID_exit:
			MemSet(&newEvent, sizeof(EventType), 0);
			newEvent.eType = appStopEvent;
			EvtAddEventToQueue(&newEvent);
			break;
	  }
      handled = 1;
      break;
	 
	case penDownEvent:
		penx=event->screenX;
		peny=event->screenY;
		if ((peny>19)&&(peny<149))
		{
			row=(peny-20)/8;
			column=penx/8;
			index=(16*column)+row;
			Select(index);
		}
		handled=1;
		break;

    case nilEvent:
      handled = 1;
      break;
  }
  return handled;
}

static Boolean infoform(EventPtr event)
{
  FormPtr   form;
  int       handled = 0;
  EventType newEvent;

  switch (event->eType) 
  {
    case frmOpenEvent:
      form = FrmGetActiveForm();
      fieldptr_ft = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_ft));
	  fieldptr_f1 = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_f1));
      fieldptr_f2 = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_f2));
      fieldptr_f3 = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_f3));
      fieldptr_f4 = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_f4));
      fieldptr_f5 = FrmGetObjectPtr(form, FrmGetObjectIndex(form, fieldID_f5));
      FrmDrawForm(form);
      GetInfo();
      handled = 1;
      break;

	case ctlSelectEvent:  // A control button was pressed and released.
      if (event->data.ctlEnter.controlID== buttonID_ok)
	  {
        FrmGotoForm(formID_game);
        handled = 1;
	  }
      break;
  
    case nilEvent:
		handled = 1;
		break;
  }
  return handled;
}

static void Select(int index)
{
	int i;
	int more=1;
	int selected=0;

	// If a new group is selected, erase previously selected group

	selected=gamearray[index];
	if (selected!=6)
	{
		for (i=0; i<320; i++)
		{
			if (gamearray[i]==6) 
			{
				gamearray[i]=selprev;
			    DrawBitmap(i);
			}
		}
	}

	if (selected==7) return;  // Cannot select white space

    // See if previously selected group is tapped again
	// Then start removing

	if (selected==6) 
	{
		StrNCopy(undoarray, gamearray, 320);
		for (i=0; i<320; i++)
		{
			if (undoarray[i]==6) undoarray[i]=selprev;
		}
		Tumble();
		return;
	}
	
	// OK, so we've got a new selection

	selprev=selected;
	gamearray[index]=6;
	DrawBitmap(index);

	while(more)
	{
		more=0;
		formv=0;
		for (i=0; i<320; i++)
		{
			if (gamearray[i]==6)
			{
				formv++;
				if (((i%16)!=15)&&(gamearray[i+1]==selected)&&(i<319))
				{
					gamearray[i+1]=6;
					DrawBitmap(i+1);
					more=1;
				}
				if (((i%16)!=0)&&(gamearray[i-1]==selected)&&(i>0))
				{
					gamearray[i-1]=6;
					DrawBitmap(i-1);
					more=1;
				}
				if ((gamearray[i+16]==selected)&&(i<304))
				{
					gamearray[i+16]=6;
					DrawBitmap(i+16);
					more=1;
				}
				if ((gamearray[i-16]==selected)&&(i>15))
				{
					gamearray[i-16]=6;
					DrawBitmap(i-16);
					more=1;
				}
			}
		}
	}

	// Cannot select only one block, so undo 

	if (formv==1) 
	{
		gamearray[index]=selected;
		DrawBitmap(index);
		points=0;
		formv=0;
	}
	if (formv>1) points=(formv*formv)-(3*formv)+4;
	UpdateScore();
}

static void GetInfo(void)
{
	int			i;
	int			t1=0, t2=0, t3=0, t4=0, t5=0;
	char		st0[5];
    char		st1[4];
    char		st2[4];
    char		st3[4];
    char		st4[4];
    char		st5[4];
    VoidHand	bitmaphandle;
    BitmapPtr	bitmap;

	for (i=0; i<320; i++)
	{
		if (gamearray[i]==1) t1++;
		if (gamearray[i]==2) t2++;
		if (gamearray[i]==3) t3++;
		if (gamearray[i]==4) t4++;
		if (gamearray[i]==5) t5++;
	}
	StrIToA(st0, topscore);
	StrIToA(st1, t1);
	StrIToA(st2, t2);
	StrIToA(st3, t3);
	StrIToA(st4, t4);
	StrIToA(st5, t5);
	FldSetTextPtr(fieldptr_ft, st0);
	FldSetTextPtr(fieldptr_f1, st1);
	FldSetTextPtr(fieldptr_f2, st2);
	FldSetTextPtr(fieldptr_f3, st3);
	FldSetTextPtr(fieldptr_f4, st4);
	FldSetTextPtr(fieldptr_f5, st5);
	FldRecalculateField(fieldptr_ft, true);
	FldRecalculateField(fieldptr_f1, true);
	FldRecalculateField(fieldptr_f2, true);
	FldRecalculateField(fieldptr_f3, true);
	FldRecalculateField(fieldptr_f4, true);
	FldRecalculateField(fieldptr_f5, true);
	FrmDrawForm(FrmGetActiveForm());
    
	bitmaphandle = DmGet1Resource('Tbmp', bitmapID_1);
    bitmap = MemHandleLock(bitmaphandle);
    WinDrawBitmap(bitmap, 48, 33);
    MemHandleUnlock(bitmaphandle);
	bitmaphandle = DmGet1Resource('Tbmp', bitmapID_2);
    bitmap = MemHandleLock(bitmaphandle);
    WinDrawBitmap(bitmap, 48, 45);
    MemHandleUnlock(bitmaphandle);
	bitmaphandle = DmGet1Resource('Tbmp', bitmapID_3);
    bitmap = MemHandleLock(bitmaphandle);
    WinDrawBitmap(bitmap, 48, 57);
    MemHandleUnlock(bitmaphandle);
	bitmaphandle = DmGet1Resource('Tbmp', bitmapID_4);
    bitmap = MemHandleLock(bitmaphandle);
    WinDrawBitmap(bitmap, 48, 69);
    MemHandleUnlock(bitmaphandle);
	bitmaphandle = DmGet1Resource('Tbmp', bitmapID_5);
    bitmap = MemHandleLock(bitmaphandle);
    WinDrawBitmap(bitmap, 48, 82);
    MemHandleUnlock(bitmaphandle);
}

static void Tumble(void)
{
	int i, j, row, column;
	int more;
	int gameover, test;
    EventType newEvent;

	// Tumble any columns with empty blocks, insert blanks at the top

	for (i=0; i<320; i++)
	{
		if (gamearray[i]==6)
		{
			column=i/16;
			row=i%16;
			for (j=row; j>0; j--)
			{
				gamearray[(column*16)+j]=gamearray[(column*16)+j-1];
				DrawBitmap((column*16)+j);
			}
			gamearray[column*16]=7;
			DrawBitmap(column*16);
		}
	}
	score=score+points;
	lastpoints=points;

	// check for empty columns, bump everything left 
	
	do 
	{
		more=0;
		for (i=0; i<20; i++)
		{
			if (gamearray[(i*16)+15]==7)
			{
				for (j=i*16; j<304; j++)
				{
					gamearray[j]=gamearray[j+16];
					if (gamearray[j]!=7) more=1;
					DrawBitmap(j);
				}
				for (j=304; j<320; j++) 
				{
					gamearray[j]=7;
					DrawBitmap(j);
				}
				break;
			}
		}
	} while (more);

	// and finally check for game over  
	// search till first two touching similar blocks are found

	gameover=1;
	for (i=0; i<320; i++)
	{
		if ((gamearray[i]>0)&&(gamearray[i]<6))
		{
			test=gamearray[i];
		
			if (((i%16)!=15)&&(gamearray[i+1]==test)&&(i<319))
			{
				gameover=0;
				break;
			}
			if (((i%16)!=0)&&(gamearray[i-1]==test)&&(i>0))
			{
				gameover=0;
				break;
			}
			if ((gamearray[i+16]==test)&&(i<303))
			{
				gameover=0;
				break;
			}
			if ((gamearray[i-16]==test)&&(i>15))
			{
				gameover=0;
				break;
			}
		}
	}
	
	if (gameover)
	{
		if (score>topscore)
		{
			topscore=score;
			FrmAlert(alertID_topscore);
		}
		GenerateArray();
		DrawGameScreen();
		score=0;
	}
	formv=0;
	points=0;
	UpdateScore();
}

static void GenerateArray(void)
{
  int			i, r;
  VoidHand      RecHandle;
  Ptr           RecPointer;

  score=0;
  for (i=0; i<320; i++)
  {
	r=1+SysRandom(0)%5;
    gamearray[i]=r;
  }
  StrNCopy(undoarray, gamearray, 320);
  StrIToA(gamearray+350, score);
  StrIToA(gamearray+360, topscore);
  points=0;
  formv=0;
  RecHandle = DmGetRecord(SameGameDB, 0);
  RecPointer = MemHandleLock(RecHandle);
  DmWrite(RecPointer, 0, &gamearray, 390);
  MemPtrUnlock(RecPointer); 
  DmReleaseRecord(SameGameDB, 0, true);
}

static void DrawBitmap(int index)
{
	VoidHand   bitmaphandle;
	BitmapPtr  bitmap;
	int column, row, rowpix, colpix;
		  
	// Calculate bitmap position
      
	column=(index-(index%16))/16;
	row=index%16;
	colpix=(column*8);
	rowpix=(row*8)+20;

	// Get correct bitmap
  
	switch (gamearray[index])
	{
		case 7:
			bitmaphandle = DmGet1Resource('Tbmp', bitmapID_0);
			break;
		case 1:
			bitmaphandle = DmGet1Resource('Tbmp', bitmapID_1);
			break;
		case 2:
			bitmaphandle = DmGet1Resource('Tbmp', bitmapID_2);
			break;
		case 3:
			bitmaphandle = DmGet1Resource('Tbmp', bitmapID_3);
			break;
		case 4:
			bitmaphandle = DmGet1Resource('Tbmp', bitmapID_4);
			break;
		case 5:
			bitmaphandle = DmGet1Resource('Tbmp', bitmapID_5);
			break;
		case 6:
			bitmaphandle = DmGet1Resource('Tbmp', bitmapID_6);
			break;
	}
   	bitmap = MemHandleLock(bitmaphandle);
	WinDrawBitmap(bitmap, colpix, rowpix);
	MemHandleUnlock(bitmaphandle);
}

static void DrawGameScreen(void)
{
	int i;

    for (i=0; i<320; i++)
	{ 
		DrawBitmap(i);
	}
}


static void UpdateScore(void)
{
  char frm[4];
  char pnts[5];

  StrIToA(frm, formv);
  StrIToA(pnts, points);
  StrIToA(gamearray+360, topscore);
  StrIToA(gamearray+350, score);
  FldSetTextPtr(fieldptr_fform, frm);
  FldSetTextPtr(fieldptr_fscore, gamearray+350);
  FldSetTextPtr(fieldptr_fpoints, pnts);
  FldRecalculateField(fieldptr_fform, true);
  FldRecalculateField(fieldptr_fscore, true);
  FldRecalculateField(fieldptr_fpoints, true);
  FrmDrawForm(FrmGetActiveForm());
}
